| C prototype: void FastFilledSquare_Draw_R(register void* plane asm("%a2"),register short x asm("%d0"),register short y asm("%d1"),register short side asm("%d2")) __attribute__((__regparm__(4)));
|
| This function ORs a rectangle to the destination. The algorithm used here
| is faster than that of FastFillRect_R. FastFillRect_R is already quite
| fast, and modifying FastFillRect_R to use this algorithm is too complicated
| and inefficient due to self-modifying code.

|addapted from FastFilledRect_Draw_R to draw a square

.data
.even

__First_FFR_D_R__:
.word 0xFFFF,0x7FFF,0x3FFF,0x1FFF,0x0FFF,0x07FF,0x03FF,0x01FF,0x00FF,0x007F,0x003F,0x001F,0x000F,0x0007,0x0003,0x0001

__Last_FFR_D_R__:
.word 0x8000,0xC000,0xE000,0xF000,0xF800,0xFC00,0xFE00,0xFF00,0xFF80,0xFFC0,0xFFE0,0xFFF0,0xFFF8,0xFFFC,0xFFFE,0xFFFF

.globl FastFilledSquare_Draw_R
.even
FastFilledSquare_Draw_R:
    movem.l  %d3-%d5/%a2,-(%sp)


    move.w %d2,%d3
    add.w %d0,%d2

| Here, we have dx in d2.
    movea.l  %a2,%a1

    move.w   %d1,%d4
    lsl.w    #4,%d1
    sub.w    %d4,%d1

    move.w   %d1,%d5   | save 15*y1 in d5

    move.w   %d0,%d4
    lsr.w    #4,%d4
    add.w    %d4,%d1   | always word-aligned to avoid address error problems.

    add.w    %d1,%d1
    adda.w   %d1,%a2
| a0=address of the top left corner of the rectangle.

    move.w   %d2,%d4
    lsr.w    #4,%d4
    add.w    %d4,%d5   | always word-aligned to avoid address error problems.

    add.w    %d5,%d5
    adda.w   %d5,%a1
| a1=address of the top right corner of the rectangle.

| Take masks of sides.
    moveq    #0xF,%d5
    and.w    %d5,%d0
    add.w    %d0,%d0
| The code was optimized and reordered so that d(pc,ix) mode can be used.
    move.w   __First_FFR_D_R__(%pc,%d0.w),%d0    | d0 = mask of first pixels.

    and.w    %d5,%d2
    add.w    %d2,%d2
    move.w   __Last_FFR_D_R__(%pc,%d2.w),%d1     | d1 = mask of last pixels.

| Are all pixels in the same word ?
    cmpa.l   %a2,%a1
    beq.s    _All_pixels_in_the_same_word_FFR_D_R__
| Short branches are faster if not taken...

| Save loop count.
    move.w   %d3,%d4

    move.l   %a2,%a0
    move.l   %a1,%d2
    sub.l    %a0,%d2
    subq.w   #1,%d2
    lsr.w    #1,%d2
| d2 = number of complete 16-pixel columns (words) between the sides.

| Mask.
    moveq    #-1,%d5

| Now, draw both sides of the rectangle.
__Draw_both_sides_FFR_D_R__:
    or.w     %d0,(%a2)
    or.w     %d1,(%a1)
    lea      30(%a2),%a2
    lea      30(%a1),%a1
    dbf      %d3,__Draw_both_sides_FFR_D_R__

| We've drawn both sides of the rectangle. Skip the left side.
    addq.l   #2,%a0

| We are going to check how many pixels we can draw at once (0, 16, 32, 64).
__How_many_pixels_can_we_write_FFR_D_R__:
| Restore loop count and pointers (after modification).
    move.w   %d4,%d3
    move.l   %a0,%a2

__Can_we_write_64_pixels_FFR_D_R__:
    subq.w   #4,%d2
    bge.s    _Write_two_cols_of_longs_FFR_D_R__
__Can_we_write_32_pixels_FFR_D_R__:
    addq.w   #2,%d2
    bge.s    __Write_one_col_of_longs_FFR_D_R__
__Can_we_write_16_pixels_FFR_D_R__:
    addq.w   #1,%d2
    bne.s    __End_FFR_D_R__
| We have one column left.

__Write_one_col_of_words_FFR_D_R__:
    move.w   %d5,(%a2)
    lea      30(%a2),%a2
    dbf      %d3,__Write_one_col_of_words_FFR_D_R__
| We've finished.
    bra.s    __End_FFR_D_R__

_Write_two_cols_of_longs_FFR_D_R__:
__Write_two_cols_of_longs_FFR_D_R__:
| Write 64 pixels at once (2*32).
    move.l   %d5,(%a2)+
    move.l   %d5,(%a2)
    lea      30-4(%a2),%a2
    dbf      %d3,__Write_two_cols_of_longs_FFR_D_R__
    addq.l   #8,%a0
    bra.s    __How_many_pixels_can_we_write_FFR_D_R__

__Write_one_col_of_longs_FFR_D_R__:
| Write 32 pixels at once (1*32). Despite the tighter loop, this is much
| more efficient size-wise and speed-wise than writing 2*16 pixels.
    move.l   %d5,(%a2)
    lea      30(%a2),%a2
    dbf      %d3,__Write_one_col_of_longs_FFR_D_R__
    addq.l   #4,%a0
    bra.s    __How_many_pixels_can_we_write_FFR_D_R__


_All_pixels_in_the_same_word_FFR_D_R__:
    and.w    %d0,%d1
__All_pixels_in_the_same_word_FFR_D_R__:
    or.w     %d1,(%a2)
    lea      30(%a2),%a2
    dbf      %d3,__All_pixels_in_the_same_word_FFR_D_R__

__End_FFR_D_R__:
    movem.l  (%sp)+,%d3-%d5/%a2
    rts
